from emmental.contrib.slicing.data import add_slice_labels
from emmental.contrib.slicing.slicing_function import slicing_function
from emmental.contrib.slicing.task import build_slice_tasks

__all__ = ["add_slice_labels", "slicing_function", "build_slice_tasks"]
